//==============================================================================
// TITLE: COpcHdaItem.cpp
//
// CONTENTS:
// 
// A collection of historical values for an item.
//
// (c) Copyright 2003-2004 The OPC Foundation
// ALL RIGHTS RESERVED.
//
// DISCLAIMER:
//  This code is provided by the OPC Foundation solely to assist in 
//  understanding and use of the appropriate OPC Specification(s) and may be 
//  used as set forth in the License Grant section of the OPC Specification.
//  This code is provided as-is and without warranty or support of any sort
//  and is subject to the Warranty and Liability Disclaimers which appear
//  in the printed OPC Specification.
//
// MODIFICATION LOG:
//
// Date       By    Notes
// ---------- ---   -----
// 2003/12/17 RSA   Initial implementation.

#include "StdAfx.h"
#include "COpcHdaItem.h"
#include "COpcHdaAttribute.h"
#include "COpcHdaHistorian.h"
#include "COpcHdaTime.h"
#include "interface.h"

//============================================================================
// Local Declarations

#define TAG_NAME       _T("Name")
#define TAG_START_TIME _T("StartTime")
#define TAG_TIMESTAMP  _T("Timestamp")

#define TICKS_PER_SECOND 10000000

//============================================================================
// COpcHdaItem

//========================================================================
// IOpcXmlSerialize


// Init
void COpcHdaItem::Init()
{
//	m_vtDataType        = VT_R8;
	m_cDescription      = _T("A description of an item.");
	m_cEngUnits         = _T("Units");
	m_cDeriveEquation   = _T("Direct");
	m_cNodeName         = _T("Localhost");
	m_cProcessName      = _T("Internal");
	m_dblNormalMax      = 1000;
	m_dblNormalMin      = 0;
	m_llMaxInterval     = 10000000;
	m_llMinInterval     = 10000000;
	m_dblDeviation      = 0;
	m_eDeviationType    = 0;
	m_dblLowEntryLimit  = DBL_MIN;
	m_dblHighEntryLimit = DBL_MAX;
	m_bLoaded           = false;

	// initialize source name and item name from the item id.
	m_cItemName = m_cItemID;

	int iIndex = m_cItemID.ReverseFind(_T("/"));

	if (iIndex != -1)
	{
		m_cItemName = m_cItemID.SubStr(iIndex+1);
	}
}
         
// Clear
void COpcHdaItem::Clear()
{
	m_cModifiedValues.RemoveAll();
	m_cAnnotations.RemoveAll();

    Init();
}

// SaveData
void COpcHdaItem::SaveData()
{
}

// Read
bool COpcHdaItem::Read(COpcXmlElement& cElement)
{
    Clear();

	return true;
}

// Write
bool COpcHdaItem::Write(COpcXmlElement& cElement)
{
	OPC_ASSERT(false);  
	
	// not implemented.
	
	return false;
}
	
// BuildAddressSpace
bool COpcHdaItem::BuildAddressSpace()
{
	if (!::GetHistorian().AddLink(m_cItemID))
	{
		return false;
	}

	return true;
}

// ClearAddressSpace
void COpcHdaItem::ClearAddressSpace()
{
	::GetHistorian().RemoveLink(m_cItemID);
}

// GetAttribute
bool COpcHdaItem::GetAttribute(DWORD dwID, OpcXml::AnyType& cValue)
{
	// store attribute value in the appropriate place.
	switch (dwID)
	{	
//		case OPCHDA_DATA_TYPE: 	        { cValue.Set((short)m_vtDataType); return true; }
		case OPCHDA_DESCRIPTION:        { cValue.Set(m_cDescription);      return true; }
		case OPCHDA_ENG_UNITS:   		{ cValue.Set(m_cEngUnits);         return true; }	
		case OPCHDA_NODE_NAME:          { cValue.Set(m_cNodeName);         return true; } 
		case OPCHDA_PROCESS_NAME:       { cValue.Set(m_cProcessName);      return true; } 
		case OPCHDA_NORMAL_MAXIMUM:     { cValue.Set(m_dblNormalMax);      return true; } 
		case OPCHDA_NORMAL_MINIMUM:     { cValue.Set(m_dblNormalMin);      return true; } 
		case OPCHDA_ITEMID:  	        { cValue.Set(m_cItemName);         return true; }	
		case OPCHDA_EXCEPTION_DEV:      { cValue.Set(m_dblDeviation);      return true; } 
		case OPCHDA_EXCEPTION_DEV_TYPE: { cValue.Set(m_eDeviationType);    return true; } 
		case OPCHDA_HIGH_ENTRY_LIMIT:   { cValue.Set(m_dblHighEntryLimit); return true; } 
		case OPCHDA_LOW_ENTRY_LIMIT:    { cValue.Set(m_dblLowEntryLimit);  return true; } 
		case OPCHDA_DERIVE_EQUATION:    { cValue.Set(m_cDeriveEquation);   return true; }	

		case OPCHDA_MAX_TIME_INT:       
		{ 
			OpcXml::Decimal decimalValue;
			decimalValue.int64 = m_llMaxInterval;
			cValue.Set(decimalValue); 
			return true; 
		}

		case OPCHDA_MIN_TIME_INT:  
		{ 
			OpcXml::Decimal decimalValue;
			decimalValue.int64 = m_llMinInterval;
			cValue.Set(decimalValue); 
			return true; 
		} 

		default:
		{
			OpcXml::AnyType* pAttribute = NULL;

			if (!m_cAttributes.Lookup(dwID, pAttribute))
			{
				return false;
			}
			
			cValue = *pAttribute;
			return true;
		}
	}
}

// Match
bool COpcHdaItem::Match(COpcHdaBrowseFilter& cFilter)
{
	// always return no match if attribute does not exist.
	OpcXml::AnyType* pValue = NULL;

	if (!m_cAttributes.Lookup(cFilter.GetID(), pValue))
	{
		return false;
	}

	OpcXml::AnyType cFilterValue(cFilter.GetValue());

	// check for type mis-match.
	if (cFilterValue.eType != pValue->eType)
	{
		return false;
	}

	// always attempt to match patterns if data type is a string.
	if (pValue->eType == OpcXml::XML_STRING && pValue->iLength < 0 && cFilterValue.iLength < 0)
	{
		return OpcMatchPattern(pValue->stringValue, cFilterValue.stringValue, true);
	}
    
	// apply the operator.
	switch (cFilter.GetOperator())
	{
		case OPCHDA_EQUAL:        { return (pValue->Compare(cFilterValue) == 0); }
		case OPCHDA_LESS:         { return (pValue->Compare(cFilterValue) <  0); }
		case OPCHDA_LESSEQUAL:    { return (pValue->Compare(cFilterValue) <= 0); }
		case OPCHDA_GREATER:      { return (pValue->Compare(cFilterValue) >  0); }
		case OPCHDA_GREATEREQUAL: { return (pValue->Compare(cFilterValue) >= 0); }
		case OPCHDA_NOTEQUAL:     { return (pValue->Compare(cFilterValue) != 0); }
	}

	// no match.
	return false;
}
// ReadModified
HRESULT COpcHdaItem::ReadModified(
	LONGLONG             llStartTime,
	LONGLONG             llEndTime,
	UINT                 uNumValues,
	UINT                 uStartIndex,
	OPCHDA_MODIFIEDITEM& cItem
)
{
	// initialize empty item object.
	cItem.dwCount             = 0;
	cItem.hClient             = NULL;
	cItem.pvDataValues        = NULL;
	cItem.pdwQualities        = NULL;
	cItem.pftTimeStamps       = NULL;
	cItem.pftModificationTime = NULL;
	cItem.pEditType           = NULL;
	cItem.szUser              = NULL;

	// return result.
	return S_OK;
}

// GetCurrentValue
HRESULT COpcHdaItem::GetCurrentValue(OPCHDA_ATTRIBUTE& cAttribute)
{
	// get the current value.
	OpcXml::AnyType cCurrentValue;

	if (!GetAttribute(cAttribute.dwAttributeID, cCurrentValue))
	{
		return OPC_E_INVALIDATTRID;
	}

	cAttribute.dwNumValues      = 1;
	cAttribute.vAttributeValues = OpcArrayAlloc(VARIANT, cAttribute.dwNumValues);
	cAttribute.ftTimeStamps     = OpcArrayAlloc(FILETIME, cAttribute.dwNumValues);

	// set timestamp.
	cAttribute.ftTimeStamps[0] = OpcUtcNow();

	// copy value.
	if (!cCurrentValue.Get(cAttribute.vAttributeValues[0]))
	{
		cAttribute.vAttributeValues[0].vt = VT_EMPTY;
		return E_FAIL;
	}

	// all done.
	return S_OK;
}


// ReadAttribute
HRESULT COpcHdaItem::ReadAttribute(
	LONGLONG          llStartTime,
	LONGLONG          llEndTime,
	DWORD             dwAttributeID, 
	OPCHDA_ATTRIBUTE& cAttribute
)
{
	// initialize empty item object.
	cAttribute.dwNumValues      = 0;
	cAttribute.dwAttributeID    = dwAttributeID;
	cAttribute.hClient          = NULL;
	cAttribute.vAttributeValues = NULL;
	cAttribute.ftTimeStamps     = NULL;

	// get the current time.
	LONGLONG llNow = OpcHdaInt64FromFILETIME(OpcUtcNow());

	// check for a request for the current value.
	if (llEndTime == 0 && ((llNow - llStartTime) < 10*TICKS_PER_SECOND))
	{
		return GetCurrentValue(cAttribute);
	}

	// determine if series is ascending or descending.
	bool bAscending = ((llStartTime != 0 && llStartTime <= llEndTime) || llEndTime == 0);

	// get lower and upper bounding times for the series.
	LONGLONG llLower = (bAscending)?llStartTime:llEndTime;
	LONGLONG llUpper = (bAscending)?llEndTime:llStartTime;

	if (llStartTime == 0)
	{
		llLower = 0;
		llUpper = llEndTime;
	}

	if (llEndTime == 0 || llUpper >= llNow)
	{
		llUpper = _I64_MAX;
	}

	Local<COpcHdaAttributeValues> attrs(GetAttributesForItem(this,llStartTime,llEndTime));

	// find indicies.
	UINT uLower = attrs->m_cValues.FindIndexAfter(llLower);
	UINT uUpper = attrs->m_cValues.FindIndexBefore(llUpper);

	// nothing after the lower bound and no upper bound - return the current value.
	if (uLower == -1 && uUpper != -1 && llUpper == _I64_MAX)
	{
		return GetCurrentValue(cAttribute);
	}

	// no historial data exists for the item. 
	if (uLower == -1 && uUpper == -1)
	{
		HRESULT hResult = GetCurrentValue(cAttribute);

		if (SUCCEEDED(hResult))
		{
			return OPC_S_CURRENTVALUE;
		}

		return hResult;
	}	

	// no data within the specified range.
	if (uUpper == -1)
	{
		return OPC_S_NODATA;
	}

	bool bNoDataAbove = (uLower == -1);

	// return value value before the lower bound.
	if (uLower == -1)
	{
		uLower = uUpper;
	}

	// count the number of values.
	COpcList<COpcHdaAttributeValue*> cValues;

	// find values within the range.
	for (UINT ii = uLower; ii <= uUpper; ii++)
	{
		if (attrs->m_cValues[ii].dwAttributeID == dwAttributeID)
		{
			cValues.AddTail(&(attrs->m_cValues[ii]));
		}
	}

	// check if it is necessary to search for an earlier value.
	bool bBoundAdded = false;

	if (cValues.GetCount() == 0 || cValues.GetHead()->llTimestamp > llLower)
	{
		while (uLower > 0)
		{
			uLower--;

			if (attrs->m_cValues[uLower].dwAttributeID == dwAttributeID)
			{
				cValues.AddHead(&(attrs->m_cValues[uLower]));
				bBoundAdded = true;
				break;
			}
		}
	}

	// determine the number of values.
	cAttribute.dwNumValues = cValues.GetCount();

	// add current value.
	if (llUpper == _I64_MAX)
	{
		cAttribute.dwNumValues++;
	}

	// no data.
	if (cAttribute.dwNumValues == 0)
	{
		// no data at all for the attribute.
		if (bNoDataAbove)
		{
			HRESULT hResult = GetCurrentValue(cAttribute);

			if (SUCCEEDED(hResult))
			{
				return OPC_S_CURRENTVALUE;
			}

			return hResult;
		}

		// no data in range.
		return OPC_S_NODATA;
	}

	// allocate arrays.
	cAttribute.vAttributeValues = OpcArrayAlloc(VARIANT, cAttribute.dwNumValues);
	cAttribute.ftTimeStamps     = OpcArrayAlloc(FILETIME, cAttribute.dwNumValues);

	memset(cAttribute.vAttributeValues, 0, sizeof(VARIANT)*cAttribute.dwNumValues);
	memset(cAttribute.ftTimeStamps, 0, sizeof(DWORD)*cAttribute.dwNumValues);

	UINT uFirst = 0;
	UINT uLast  = cAttribute.dwNumValues-1;

	// add current value.
	if (llUpper == _I64_MAX)
	{
		cAttribute.ftTimeStamps[uLast] = OpcUtcNow();

		OpcXml::AnyType cCurrentValue;

		if (GetAttribute(cAttribute.dwAttributeID, cCurrentValue))
		{
			cCurrentValue.Get(cAttribute.vAttributeValues[uLast]);
		}

		if (uLast > 0)
		{
			uLast--;
		}
	}

	// fill arrays.
	while (uFirst <= uLast && cValues.GetCount() > 0)
	{
		COpcHdaAttributeValue* pValue = (bAscending)?cValues.RemoveHead():cValues.RemoveTail();
		
		// adjust timestamps up to the lower bound.
		if (pValue->llTimestamp < llLower)
		{
			cAttribute.ftTimeStamps[uFirst] = OpcHdaFILETIMEFromInt64(llLower);
		}
		else
		{
			cAttribute.ftTimeStamps[uFirst] = OpcHdaFILETIMEFromInt64(pValue->llTimestamp);
		}

		// copy value.
		VariantCopy(&cAttribute.vAttributeValues[uFirst], pValue->cValue.GetPtr());

		uFirst++;
	}

	// all done.
	return S_OK;
}

// ReadAnnotations
HRESULT COpcHdaItem::ReadAnnotations(
	LONGLONG           llStartTime,
	LONGLONG           llEndTime,
	OPCHDA_ANNOTATION& cAnnotation
)
{
	// initialize empty item object.
	cAnnotation.dwNumValues      = 0;
	cAnnotation.hClient          = NULL;
	cAnnotation.ftTimeStamps     = NULL;
	cAnnotation.szAnnotation     = NULL;
	cAnnotation.ftAnnotationTime = NULL;
	cAnnotation.szUser           = NULL;

	// determine if series is ascending or descending.
	bool bAscending = ((llStartTime != 0 && llStartTime <= llEndTime) || llEndTime == 0);

	// get lower and upper bounding times for the series.
	LONGLONG llLower = (bAscending)?llStartTime:llEndTime;
	LONGLONG llUpper = (bAscending)?llEndTime:llStartTime;

	if (llStartTime == 0)
	{
		llLower = 0;
		llUpper = llEndTime;
	}

	if (llEndTime == 0)
	{
		llUpper = _I64_MAX;
	}

	// find indicies.
	UINT uLower = m_cAnnotations.FindIndexAfter(llLower);
	UINT uUpper = m_cAnnotations.FindIndexBefore(llUpper);

	if (uLower == -1 || uUpper == -1 || uLower > uUpper)
	{
		return OPC_S_NODATA;
	}

	HRESULT hResult = S_OK;

	// calculate number of values to return.
	cAnnotation.dwNumValues = uUpper - uLower + 1;

	// allocate arrays.
	cAnnotation.ftTimeStamps     = OpcArrayAlloc(FILETIME, cAnnotation.dwNumValues);
	cAnnotation.szAnnotation     = OpcArrayAlloc(LPWSTR, cAnnotation.dwNumValues);
	cAnnotation.ftAnnotationTime = OpcArrayAlloc(FILETIME, cAnnotation.dwNumValues);
	cAnnotation.szUser           = OpcArrayAlloc(LPWSTR, cAnnotation.dwNumValues);

	memset(cAnnotation.ftTimeStamps, 0, sizeof(FILETIME)*cAnnotation.dwNumValues);
	memset(cAnnotation.szAnnotation, 0, sizeof(LPWSTR)*cAnnotation.dwNumValues);
	memset(cAnnotation.ftAnnotationTime, 0, sizeof(FILETIME)*cAnnotation.dwNumValues);
	memset(cAnnotation.szUser, 0, sizeof(LPWSTR)*cAnnotation.dwNumValues);

	// fill arrays.
	for (UINT ii = 0; ii < cAnnotation.dwNumValues; ii++)
	{
		COpcHdaAnnotation& cValue = m_cAnnotations[((bAscending)?uLower+ii:uUpper-ii)];

		cAnnotation.ftTimeStamps[ii]     = OpcHdaFILETIMEFromInt64(cValue.llTimestamp);
		cAnnotation.szAnnotation[ii]     = OpcStrDup((LPCWSTR)cValue.cAnnotation);
		cAnnotation.ftAnnotationTime[ii] = OpcHdaFILETIMEFromInt64(cValue.llCreationTime);
		cAnnotation.szUser[ii]           = OpcStrDup((LPCWSTR)cValue.cUser);
	}

	// return result.
	return hResult;
}

// InsertAnnotations
HRESULT COpcHdaItem::InsertAnnotation(
	LONGLONG           llTimestamp,
	OPCHDA_ANNOTATION& cAnnotation
)
{
	for (DWORD ii = 0; ii < cAnnotation.dwNumValues; ii++)
	{
		// add new annotation.
		COpcHdaAnnotation cNewValue;

		cNewValue.llTimestamp    = llTimestamp;
		cNewValue.cAnnotation    = cAnnotation.szAnnotation[ii];
		cNewValue.llCreationTime = OpcHdaInt64FromFILETIME(cAnnotation.ftAnnotationTime[ii]);
		cNewValue.cUser          = cAnnotation.szUser[ii];

		m_cAnnotations.Insert(llTimestamp, cNewValue);
	}

	// all done.
	return S_OK;
}

// CreateModifiedValue
void COpcHdaItem::CreateModifiedValue(OPCHDA_EDITTYPE eEditType, COpcHdaItemValue& cValue)
{
	// create modified value.
	COpcHdaModifiedValue cModifiedValue;

	cModifiedValue.value           = cValue.value;
	cModifiedValue.dwQuality          = cValue.dwQuality;
	cModifiedValue.llTimestamp        = cValue.llTimestamp;
	cModifiedValue.llModificationTime = OpcHdaInt64FromFILETIME(OpcUtcNow());
	cModifiedValue.eEditType          = eEditType;

	// lookup user name.
	TCHAR tsUserName[_MAX_PATH+1];
	memset(tsUserName, 0, sizeof(tsUserName));

	DWORD dwLength = _MAX_PATH;

	if (GetUserName(tsUserName, &dwLength))
	{
		cModifiedValue.cUser =  tsUserName;
	}

	// update modified value table.
	m_cModifiedValues.Insert(cModifiedValue.llTimestamp, cModifiedValue);
}

// Update
HRESULT COpcHdaItem::Update(
	OPCHDA_EDITTYPE eEditType,
	LONGLONG        llTimestamp,
	VARIANT&        vValue,
	DWORD           dwQuality
)
{
	// all done.
	return S_OK;
}

// Delete
HRESULT COpcHdaItem::Delete(
	LONGLONG llStartTime,
	LONGLONG llEndTime
)
{
	// all done.
	return S_OK;
}