#ifndef _COpcHdaItemValues_H_
#define _COpcHdaItemValues_H_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "COpcHdaAttribute.h"
#include "COpcHdaItem.h"
#include "COpcHdaItemValue.h"
#include "COpcXmlElement.h"
#include "OpcDefs.h"
#include "opchda.h"
#include "libopchda.h"

//==============================================================================
// CLASS:   COpcHdaItem
// PURPOSE: A source server for data exchange data.

class COpcHdaItemValues : public IOpcXmlSerialize
{
    OPC_CLASS_NEW_DELETE()

public:

    //==========================================================================
    // Operators

    // Constructor
    COpcHdaItemValues(COpcHdaItem *item,int count);

    // Destructor 
    ~COpcHdaItemValues() { Clear(); }
    
	// reads the raw data for the specified interval.
	HRESULT ReadRaw(
		LONGLONG     llStartTime,
		LONGLONG     llEndTime,
		UINT         uNumValues,
		bool         bIncludeBounds,
		UINT         uStartIndex,
		OPCHDA_ITEM& cItem
	);

	// reads the processed data for the specified interval.
	HRESULT ReadProcessed(
		LONGLONG     llStartTime,
		LONGLONG     llEndTime,
		LONGLONG     llResampleInterval,
		DWORD        haAggregate, 
		OPCHDA_ITEM& cItem,
		bool         useEndBound
	);

	// reads the data at the specified times.
	HRESULT ReadAtTime(
		COpcArray<LONGLONG>& cTimestamps,
		OPCHDA_ITEM&         cItem
	);

    // Init
    virtual void Init();

    // Clear
    virtual void Clear();

    // Read
    virtual bool Read(COpcXmlElement& cElement);

    // Write
    virtual bool Write(COpcXmlElement& cElement);
	
	void AppendValue(COpcVariant &value,LONGLONG when, DWORD dwQuality);

	void FixProcessing(LONGLONG start,LONGLONG end);

private:
    
	// checks whether the value at index has good quality.
	bool IsUseableData(UINT uIndex);
	bool IsUseableData(const COpcHdaItemValue& cValue);

	// gets the timestamp of the sample at the specified index.
	LONGLONG GetTimestampAtIndex(UINT uIndex);

	// returns the value at the specified time by interpolating between the bounding values.
	double Interpolate(LONGLONG llTime, DWORD& dwQuality);

	// calculates the time average by interpolating the values at the start and end.
	double TimeAverage(LONGLONG llStartBound, LONGLONG llEndBound, DWORD& dwQuality);

	// calculates the total area under the curve.
	double Total(LONGLONG llStartBound, LONGLONG llEndBound, DWORD& dwQuality);

	// calculates the arthimatic average of all good values in the interval.
	double Average(LONGLONG llStartBound, LONGLONG llEndBound, DWORD& dwQuality);

	// counts the number of good values in the interval.
	double Count(LONGLONG llStartBound, LONGLONG llEndBound, DWORD& dwQuality);

	// calculates the standard deviation for the set of values in the interval.
	double StdDev(bool bVariance, LONGLONG llStartTime, LONGLONG llEndTime, DWORD& dwQuality);
		
	// finds the minimum or maximum good value in the interval.
	double MinMax(
		bool      bMinimum,
		LONGLONG  llStartTime, 
		LONGLONG  llEndTime, 
		DWORD&    dwQuality,
		FILETIME* pftTimestamp
	);

	// finds the start or end value within the interval.
	double StartEnd(
		bool      bStart,
		LONGLONG  llStartTime, 
		LONGLONG  llEndTime, 
		DWORD&    dwQuality,
		FILETIME& ftTimestamp
	);

	// calculates the difference between the first and last good values in the interval.
	double Delta(LONGLONG llStartTime, LONGLONG llEndTime, DWORD& dwQuality);
	
	// calculates the absolute difference between the maximum and minimum values in the interval.
	double Range(LONGLONG llStartTime, LONGLONG llEndTime, DWORD& dwQuality);

	// calculates a linear regression for all points in the interval.
	double Regression(
		bool      bSlope,
		LONGLONG  llStartTime, 
		LONGLONG  llEndTime, 
		DWORD&    dwQuality
	);

	// calculates the error term for a linear regression for all points in the interval.
	double RegressionDeviation(
		LONGLONG  llStartTime, 
		LONGLONG  llEndTime, 
		DWORD&    dwQuality
	);

	// calculates a linear regression on all points in the interval.
	DWORD LinearRegression(
		LONGLONG llStartTime, 
		LONGLONG llEndTime, 
		double&  dblSlope,
		double&  dblConstant
	);

	// calculates the time or percentage of an interval that is either good/bad.
	double Duration(
		bool      bGood,
		bool      bPercent,
		LONGLONG  llStartTime, 
		LONGLONG  llEndTime, 
		DWORD&    dwQuality
	);

	// returns the worst quality in the interval.
	double WorstQuality(LONGLONG llStartTime, LONGLONG llEndTime, DWORD& dwQuality);

	// finds all values within the specified interval. returns false if no data exists.
	bool FindRawValues(LONGLONG llStartTime, LONGLONG llEndTime, COpcList<COpcHdaItemValue*>& cValues);

	// finds all usuable values within the specified interval. returns the uncertain quality if bad data exists.
	DWORD FindUseableValues(LONGLONG llStartTime, LONGLONG llEndTime, COpcList<COpcHdaItemValue*>& cValues);

	// fills an array with the indicies of the selected values.
	bool ReadIndicies(
		LONGLONG llStartTime,
		LONGLONG llEndTime,
		UINT     uMaxValues,
		bool     bIncludeBounds,
		UINT*    pIndicies,
		UINT&    uCount
	);

	VARIANT GetValueIndex(UINT uIndex, DWORD& dwQuality);
	VARIANT GetValueVariant(COpcVariant & value, DWORD& dwQuality);
	VARIANT GetValueDouble(double value, DWORD& dwQuality);
	VARIANT GetValueAtTime(LONGLONG when, DWORD& dwQuality);

	// updates the start/end times after data was modified.
	void FixBoundingTimes();

    //==========================================================================
    // Private Members

	// identity attributes.
	COpcHdaItem *item;

	double     m_dblNormalMax;
	double     m_dblNormalMin;
	LONGLONG   m_llMaxInterval;
	LONGLONG   m_llMinInterval;
	double     m_dblDeviation;
	short      m_eDeviationType;
	double     m_dblLowEntryLimit;
	double     m_dblHighEntryLimit;

	// data processing attributes.
	bool       m_bArchiving;
	bool       m_bStepped;
	UINT       m_uWaveform;
	double     m_dblPeriod;

    LONGLONG                   m_llStartTime;
	LONGLONG                   m_llEndTime;
	UINT                       m_uFirstUseable;
	UINT                       m_uLastUseable;

	COpcHdaItemValueArray      m_cValues;
};


class COpcHdaAttributeValues
{
    OPC_CLASS_NEW_DELETE_ARRAY()

public:

	void Init();
	void Clear();

    // Constructor
    COpcHdaAttributeValues(COpcHdaItem *item,int count);

    // Destructor 
    ~COpcHdaAttributeValues() { Clear(); }
    
	void AppendValue(COpcVariant &value,LONGLONG when, DWORD dwAttributeID);

	// Parse
	COpcHdaItem *item;
	COpcHdaAttributeValueArray m_cValues;
};


#endif // _COpcHdaItemValues_H_
